﻿Imports System.Runtime.CompilerServices

Namespace System.Collections.Generic
   Public Module IEnumeratorX
      ''' <summary> wrapper- **class** around Struct-IEnumerators, which would fail with .Yield() etc. Because (.Yield)-Moving a (struct)-Enumerator-Copy doesn't move the original Enumerator </summary>
      Private Class _Enumerator(Of T) : Implements IEnumerator(Of T)
         Private _Enrt As IEnumerator(Of T)
         Public ReadOnly Property Enrt() As IEnumerator(Of T)
            Get
               Return _Enrt
            End Get
         End Property
         Public Sub New(Enrt As IEnumerator(Of T))
            _Enrt = Enrt
         End Sub
         Public ReadOnly Property Current As T Implements IEnumerator(Of T).Current
            Get
               Return _Enrt.Current
            End Get
         End Property
         Private ReadOnly Property Current1 As Object Implements IEnumerator.Current
            Get
               Return Current
            End Get
         End Property
         Public Function MoveNext() As Boolean Implements IEnumerator.MoveNext
            Return _Enrt.MoveNext
         End Function
         Public Sub Reset() Implements IEnumerator.Reset
            _Enrt.Reset()
         End Sub
         Public Sub Dispose() Implements IDisposable.Dispose
            _Enrt.Dispose()
         End Sub
      End Class

      ''' <summary> ensures to get an Enumerator-**Class** to support proper work of .Yield()-Extensions etc. </summary>
      <DebuggerStepThrough(), Extension()> _
      Public Function GetEnumeratorX(Of T)(enmbl As IEnumerable(Of T)) As IEnumerator(Of T)
         Dim enrt = enmbl.GetEnumerator
         If TypeOf enrt Is ValueType Then Return New _Enumerator(Of T)(enrt)
         Return enrt
      End Function

      <DebuggerStepThrough(), Extension()> _
      Public Function Skip(Of T)(enrt As IEnumerator(Of T), Optional count As Integer = 1) As IEnumerator(Of T)
         For i = 0 To count - 1
            If Not enrt.MoveNext Then Exit For
         Next
         Return enrt
      End Function

      <DebuggerStepThrough(), Extension()> _
      Public Iterator Function TakeUntil(Of T)(enrt As IEnumerator(Of T), condition As Func(Of T, Boolean)) As IEnumerable(Of T)
         While enrt.MoveNext() AndAlso Not condition(enrt.Current)
            Yield enrt.Current
         End While
      End Function

      <DebuggerStepThrough(), Extension()> _
      Public Iterator Function TakeWhile(Of T)(enrt As IEnumerator(Of T), condition As Func(Of T, Boolean)) As IEnumerable(Of T)
         While enrt.MoveNext() AndAlso condition(enrt.Current)
            Yield enrt.Current
         End While
      End Function

      <DebuggerStepThrough(), Extension()> _
      Public Function Yield(Of T)(enrt As IEnumerator(Of T)) As T
         enrt.MoveNext()
         Return enrt.Current
      End Function
      <DebuggerStepThrough(), Extension()> _
      Public Iterator Function YieldAllToEnd(Of T)(enrt As IEnumerator(Of T)) As IEnumerable(Of T)
         While enrt.MoveNext() : Yield enrt.Current : End While
      End Function

      ''' <summary>Exception on empty collection</summary>
      <DebuggerStepThrough(), Extension()> _
      Public Function YieldCyclic(Of T As Class)(enrt As IEnumerator(Of T)) As T
         With enrt
            If .MoveNext Then Return .Current
            .Reset() : .MoveNext() : Return .Current
         End With
      End Function

      <DebuggerStepThrough(), Extension()> _
      Public Function Yield(enrt As IEnumerator) As Object
         enrt.MoveNext()
         Return enrt.Current
      End Function
      <DebuggerStepThrough(), Extension()> _
      Public Function All(Of T)(enrt As IEnumerator(Of T), condition As Func(Of T, Boolean)) As Boolean
         While enrt.MoveNext
            If Not condition(enrt.Current) Then Return False
         End While
         Return True
      End Function
      <DebuggerStepThrough(), Extension()> _
      Public Function Any(Of T)(enrt As IEnumerator(Of T), condition As Func(Of T, Boolean)) As Boolean
         While enrt.MoveNext
            If condition(enrt.Current) Then Return True
         End While
         Return False
      End Function
      <DebuggerStepThrough(), Extension()> _
      Public Sub ForEach(Of T)(enrt As IEnumerator(Of T), Action As Action(Of T))
         While enrt.MoveNext : Action(enrt.Current) : End While
      End Sub

   End Module
End Namespace