﻿#Region "FileHeader"
#If False Then
Das <Extension>-Attribute bestimmt, dass der 1. Parameter wie ein Objekt anzusprechen ist.
#End If
Imports System.Runtime.CompilerServices
#End Region 'FileHeader

Public Module DataExtensions

   ''' <summary> fügt eine gültige neue Row hinzu, inklusive der erforderlichen Verweise auf ParentRows - ausser Key-Spalten werden die Werte von rwSrc überspielt </summary>
   <Extension>
   Public Function ImportRowX(Of T As DataRow)(tb As TypedTableBase(Of T), rwSrc As T, ParamArray parentRows As DataRow()) As T
      Return DirectCast(DirectCast(tb, DataTable).ImportRowX(rwSrc, parentRows), T)
   End Function
   <Extension>
   Public Function All(Of T As DataRow)(tb As T()) As IEnumerable(Of T)
      Return From rw In tb Where 0 = (rw.RowState And (DataRowState.Deleted Or DataRowState.Detached))
   End Function
   ''' <summary> fügt eine gültige neue Row hinzu, inklusive der erforderlichen Verweise auf ParentRows - ausser Key-Spalten werden die Werte von rwSrc überspielt </summary>
   <Extension>
   Public Function ImportRowX(tb As DataTable, rwSrc As DataRow, ParamArray parentRows As DataRow()) As DataRow
      Dim rw = tb.NewRow
      Dim iPrims = tb.PrimaryKey.ToArray(Function(col) col.Ordinal)
      Dim itms = rwSrc.ItemArray
      iPrims.ForEach(Sub(i) itms(i) = rw(i))
      rw.ItemArray = itms
      If parentRows.Length > 0 Then SetParents(rw, parentRows, tb.GetAssociatedRelations(parentRows, False))
      tb.Rows.Add(rw)
      Return rw
   End Function

   ''' <summary> fügt eine gültige neue Row hinzu, inklusive der erforderlichen Verweise auf ParentRows </summary>
   <Extension>
   Public Function AddDefaultRow(Of T As DataRow)(tb As TypedTableBase(Of T), ParamArray parentRows As DataRow()) As T
      Dim rw = tb.NewRow
      If parentRows.Length > 0 Then SetParents(rw, parentRows, tb.GetAssociatedRelations(parentRows, False))
      tb.Rows.Add(rw)
      Return DirectCast(rw, T)
   End Function

   ''' <summary> legt in allen Spalten wo's fehlt Standard-DefaultValues fest </summary>
   <Extension> _
   Public Sub EnsureDefaults(Table As DataTable)
      For Each col As DataColumn In Table.Columns
         With col
            'nicht nötig bei AllowDbNull, PrimKeys, AutoIncrement oder wenns den DefaultValue schon gibt
            If .AllowDBNull OrElse .Unique OrElse .AutoIncrement OrElse Not Convert.IsDBNull(.DefaultValue) Then Continue For
            If .DataType Is GetType(String) Then
               .DefaultValue = "<n.a.>"
            Else
               .DefaultValue = Activator.CreateInstance(.DataType)
            End If
         End With
      Next
   End Sub

   <Extension> _
   Public Function IsDefault(clm As DataColumn, rw As DataRow) As Boolean
      Return Object.Equals(rw(clm), clm.DefaultValue)
   End Function

   Private Function GetCommonChildRows(parentRows As DataRow(), relations As DataRelation()) As DataRow()
      Using enrt = parentRows.GetEnumeratorX, rels = relations.GetEnumeratorX
         With enrt
            Dim candidates = .Yield.GetChildRows(rels.Yield)
            While .MoveNext
               candidates = candidates.Intersect(.Current.GetChildRows(rels.Yield)).ToArray
               If candidates.Length = 0 Then Exit While
            End While
            Return candidates
         End With
      End Using
   End Function

   Private _AmbiguousException As New System.Reflection.AmbiguousMatchException("Mindestens einer parentrow könnte mehr als eine DataRelation zugeordnet werden")
   <Extension>
   Private Function GetAssociatedRelations(tbChild As DataTable, parentRows As DataRow(), throwOnAmbiguous As Boolean) As DataRelation()
      'zu 'throwOnAmbiguous': Bestehen mehrere DataRelations zw. 2 Tabellen, so kann die betroffene parentRow nicht eindeutig zugeordnet werden. Das ist unproblematisch beim Generieren von Datensätzen, aber Suchen anhand dieser parentRows wären uneindeutig 
      Dim rels = tbChild.ParentRelations.Cast(Of DataRelation).ToArray
      Dim hsh = New HashSet(Of DataTable)
      For i = 0 To parentRows.Length - 1
         Dim tbl = parentRows(i).Table
         If throwOnAmbiguous AndAlso Not hsh.Add(tbl) Then Throw _AmbiguousException
         Dim idx = rels.FindIndex(Function(rl) rl.ParentTable Is tbl, i)
         Select Case idx
            Case i
            Case Is < 0 : Throw New InvalidOperationException("Es konnte keine (bei Doubletten: weitere) Relation {}->{} gefunden werden.".FormatX(tbl.TableName, tbChild.TableName))
            Case Else : rels.Swap(i, idx) 'rel nach vorn tauschen
         End Select
      Next
      If parentRows.Length < rels.Length Then
         If throwOnAmbiguous AndAlso hsh.Overlaps(rels.Skip(parentRows.Length).Select(Function(rl) rl.ParentTable)) Then Throw _AmbiguousException
         ReDim Preserve rels(parentRows.Length - 1)
      End If
      Return rels
   End Function

   Private Sub SetParents(rw As DataRow, parentRows As DataRow(), parentRels As DataRelation())
      For i = 0 To parentRows.Length - 1
         rw.SetParentRow(parentRows(i), parentRels(i))
      Next
   End Sub

   Public Class TableFind(Of T As DataRow)

#If False Then
Ein TableFind findet keine, eine oder mehrere Rows. Kriterium kann eine Spalte und SpaltenWert sein, und/oder ggfs. mehrere ParentRows (Schnittmenge derer ChildRows). Wird die Spalte ohne Wert angegeben, so wird nach deren DefaultWert gesucht, bzw. nach DbNull.
Der CType-Operator ermöglicht, den TableFind auch direkt an eine typDataRow-Variable zuzuweisen - es wird dabei der Single-Match (oder Nothing) zugewiesen.
Bei mehrere Matches Exception, es sei denn, das FindAny-Flag ist gesetzt.
TableFind speichert die Suchkriterien, und kann - mit .OrCreate() - eine entsprechende Row bei NichtFinden auch generieren. Hierbei können weitere ParentRows angegeben werden, welche zur Erstellung erforderlich sind, die aber kein Suchkriterium sein sollen.
Mit den überladenen Extensions Find() und FindAny() steht somit ein mächtiges Instrument bereit, um kompakt und differenziert zu suchen und im Fehlen-Fall zu erzeugen - Beispiele:

Dim rwMesstag = MessTag.Find(MessTag.DatumColumn, datum, rwWindpark).OrCreate(rwPrufer1, rwPrufer2)
sucht/erzeugt in Messtag nach einer rwWindpark untergeordneten Row mit angegebenem Datum. Beim Erzeugen werden Verweise auf die beiden Prüfer gesetzt (ansonsten wäre die neue Row ungültig (ForeignKeyConstraint))

Dim rwMessreihe = MessReihe.Find(rwWea, rwMesstag).OrCreate(rwWitterung)
Messreihe der Wea am Messtag suchen oder createn - beim createn auf Witterung verweisen.

Dim fnd = Windpark.Find(Windpark.NameColumn, sWindpark)
Dim rwWindpark = fnd.OrCreate
If Not fnd.MatchFound Then ExternWindpark.AddExternWindparkRow(rwWindpark)
fnd sucht/erzeugt einen Windpark mit angegebenem Namen. Wurde der Windpark erst erzeugt, findet eine weitere Aktion statt
#End If

      Private _Table As TypedTableBase(Of T), _Column As DataColumn, _Value As Object, _ParentRows As DataRow(), _ParentRels As DataRelation()
      Public ReadOnly MatchFound, FindAny As Boolean, Matches As T()

      ''' <summary> bei value=Nothing wird der DefaultValue der column als SuchPattern genommen </summary>
      Public Sub New(table As TypedTableBase(Of T), column As DataColumn, value As Object, findAny As Boolean, parentRows As DataRow())
         _Table = If(table, DirectCast(column.Table, TypedTableBase(Of T)))
         If _Table.Null Then Throw Me.Exception("entweder table oder column muss angegeben sein!")
         _Column = column : _Value = value : Me.FindAny = findAny : _ParentRows = If(parentRows, {})
         Dim mts As DataRow() = Nothing
         If _ParentRows.Length = 0 Then
            mts = table.AsEnumerable.ToArray
         Else
            _ParentRels = _Table.GetAssociatedRelations(_ParentRows, True)
            mts = GetCommonChildRows(parentRows, _ParentRels)
         End If
         With _Column
            If .NotNull Then
               If _Value Is Nothing Then _Value = .DefaultValue
               If Convert.IsDBNull(_Value) Then
                  If Not .AllowDBNull Then
                     If .DataType Is GetType(String) Then
                        _Value = "<n.a.>"
                     Else
                        _Value = Activator.CreateInstance(.DataType)
                     End If
                  End If
               End If
               If .AllowDBNull Then
                  mts = mts.Where(Function(rw) Not rw.IsNull(_Column) AndAlso Object.Equals(_Value, rw(_Column))).ToArray
               Else
                  mts = mts.Where(Function(rw) Object.Equals(_Value, rw(_Column))).ToArray
               End If
            End If
         End With
         Matches = mts.ToArray(Function(rw) DirectCast(rw, T))
         MatchFound = Matches.Length > 0
      End Sub

      Public Function OrCreate(callBack As Action(Of T), ParamArray parentRows As DataRow()) As T
         If MatchFound Then Return Match
         Dim rw = _Table.NewRow
         If _ParentRows.Length > 0 Then SetParents(rw, _ParentRows, _ParentRels)
         If parentRows.Length > 0 Then SetParents(rw, parentRows, _Table.GetAssociatedRelations(parentRows, False))
         If _Column.NotNull Then rw(_Column) = _Value
         OrCreate = DirectCast(rw, T)
         callBack(OrCreate)
         _Table.Rows.Add(rw)
      End Function

      Public Function OrCreate(ParamArray parentRows As DataRow()) As T
         If MatchFound Then Return Match
         Dim rw = _Table.NewRow
         If _ParentRows.Length > 0 Then SetParents(rw, _ParentRows, _ParentRels)
         If parentRows.Length > 0 Then SetParents(rw, parentRows, _Table.GetAssociatedRelations(parentRows, False))
         If _Column.NotNull Then rw(_Column) = _Value
         _Table.Rows.Add(rw)
         Return DirectCast(rw, T)
      End Function

      Public ReadOnly Property Match As T
         Get
            Return If(FindAny, Matches.FirstOrDefault, Matches.SingleOrDefault)
         End Get
      End Property

      Public Shared Widening Operator CType(this As TableFind(Of T)) As T
         Return this.Match
      End Operator

   End Class 'TableFind

   ''' <summary> sucht die einzige Row, deren identCol identValue aufweist - sowie die angegebenen ParentRows </summary>
   <Extension> _
   Public Function Find(Of T As DataRow)(tbl As TypedTableBase(Of T), identCol As DataColumn, identValue As Object, ParamArray parentRows As DataRow()) As TableFind(Of T)
      Return New TableFind(Of T)(tbl, identCol, identValue, False, parentRows)
   End Function
   ''' <summary> sucht die einzige Row, deren identCol ihren Defaultvalue aufweist - sowie die angegebenen ParentRows </summary>
   <Extension> _
   Public Function Find(Of T As DataRow)(tbl As TypedTableBase(Of T), identCol As DataColumn, ParamArray parentRows As DataRow()) As TableFind(Of T)
      Return New TableFind(Of T)(tbl, identCol, Nothing, False, parentRows)
   End Function
   ''' <summary> sucht die einzige Row mit den angegebenen parentRows </summary>
   <Extension> _
   Public Function Find(Of T As DataRow)(tbl As TypedTableBase(Of T), ParamArray parentRows As DataRow()) As TableFind(Of T)
      Return New TableFind(Of T)(tbl, Nothing, Nothing, False, parentRows)
   End Function

   ''' <summary> sucht die erste Row, deren identCol identValue aufweist - sowie die angegebenen ParentRows </summary>
   <Extension> _
   Public Function FindAny(Of T As DataRow)(tbl As TypedTableBase(Of T), identCol As DataColumn, identValue As Object, ParamArray parentRows As DataRow()) As TableFind(Of T)
      Return New TableFind(Of T)(tbl, identCol, identValue, True, parentRows)
   End Function
   ''' <summary> sucht die erste Row, deren identCol ihren Defaultvalue aufweist - sowie die angegebenen ParentRows </summary>
   <Extension> _
   Public Function FindAny(Of T As DataRow)(tbl As TypedTableBase(Of T), identCol As DataColumn, ParamArray parentRows As DataRow()) As TableFind(Of T)
      Return New TableFind(Of T)(tbl, identCol, Nothing, True, parentRows)
   End Function
   ''' <summary> sucht die erste Row mit den angegebenen parentRows </summary>
   <Extension> _
   Public Function FindAny(Of T As DataRow)(tbl As TypedTableBase(Of T), ParamArray parentRows As DataRow()) As TableFind(Of T)
      Return New TableFind(Of T)(tbl, Nothing, Nothing, True, parentRows)
   End Function

End Module
