﻿Imports System.ComponentModel
Imports System.Collections.ObjectModel
Imports System.Reflection
Imports System.Collections.Specialized
Imports Bestellungen.BestellungenEdm
Imports System.Data.Entity

Namespace Viewmodel
   Public Class Mainmodel : Inherits NotifyPropertyChanged
      Public Shared Instance As New Mainmodel

      Public WithEvents Commands As New CommandTree

      Private _Context As BestellungenEdmContainer

      Public WithEvents Kunden As ListCollectionView
      Public WithEvents Artikel As ListCollectionView
      Public WithEvents Bestellungen As ListCollectionView
      Public WithEvents BestellPosten As ListCollectionView

      Private Sub New()
         Commands("").AddFlow(AddressOf GenerateData)(AddressOf Reload)(AddressOf Save, AddressOf CanSave)("Log Sql")(AddressOf Test).Nop() ' .Nop(): Flow-Indexer erzwingen eine Auswertung
         If IsInDesignMode Then GenerateData() : Return
         _Context = New BestellungenEdmContainer
         AddHandler Commands("3").CheckedChanged, Sub(s, e) SetLogSql()
         Commands("3").SetChecked(True)
         Dim db = _Context.Database
         Dim builder = New System.Data.SqlClient.SqlConnectionStringBuilder(db.Connection.ConnectionString)
         Dim file = builder.AttachDBFilename.Replace("|DataDirectory|", AppDomain.CurrentDomain.GetData("DataDirectory").ToString & "\")
         Dbg("AttachedDBFile: {}".FormatX(file))
         Dbg("Database new created: {}".FormatX(db.CreateIfNotExists))
      End Sub
      
      Private Sub GenerateData()
         Dim kundn = New List(Of Kunde)(10.Times(Function(i) New Kunde With {.Name = "Kunde_" & i}))
         Dim artikl = New List(Of Artikel)(10.Times(Function(i) New Artikel With {.Name = "Artikel_" & i, .Preis = (i + 5)}))
         Dim dt = Date.Today.AddDays(-600)
         Dim rnd As New Random(999)
         For Each kd In kundn
            For i = 0 To 9
               Dim best = New Bestellung With {.Datum = dt.AddDays(i), .Kunde = kd}
               Call 10.Times(Sub(ii) best.BestellPosten.Add(New BestellPosten With {.Artikel = artikl(ii), .Count = rnd.Next(1, 50)}))
               kd.Bestellung.Add(best)
            Next
            dt = dt.AddDays(50)
         Next
         If IsInDesignMode Then
            Kunden = New ListCollectionView(kundn)
            Artikel = New ListCollectionView(artikl)
         Else
            Kunden = _Context.CreateRegisteredCollectionView(kundn)
            Artikel = _Context.CreateRegisteredCollectionView(artikl)
            Commands("3").SetChecked(False)
            RaisePropChanged("Kunden", "Artikel")
         End If
         Call {Kunden, Artikel}.ForEach(Sub(cv) cv.MoveCurrentToFirst())
      End Sub

      Private Sub Kunden_CurrentChanged(sender As Object, e As EventArgs) Handles Kunden.CurrentChanged
         _Context.ExchangeChildCV(Kunden, Bestellungen, Function(itm As Kunde) itm.Bestellung)
         RaisePropChanged("Bestellungen")
      End Sub

      Private Sub Bestellungen_CurrentChanged(sender As Object, e As EventArgs) Handles Bestellungen.CurrentChanged
         _Context.ExchangeChildCV(Bestellungen, BestellPosten, Function(itm As Bestellung) itm.BestellPosten)
         RaisePropChanged("BestellPosten")
      End Sub

      Private Sub Reload()
         _Context.Dispose()
         _Context = New BestellungenEdmContainer
         SetLogSql()
         Artikel = _Context.CreateRegisteredCollectionView(_Context.ArtikelSet)
         Kunden = _Context.CreateRegisteredCollectionView(_Context.KundeSet)
         RaisePropChanged("Kunden", "Artikel")
      End Sub

      Private Sub SetLogSql()
         _Context.Database.Log = If(Commands("3").IsChecked, DirectCast(AddressOf Console.Write, Action(Of String)), Nothing)
      End Sub

      Private Sub Save()
         Dbg("{} records saved".FormatX(_Context.SaveChanges()))
      End Sub

      Private Function CanSave() As Boolean
         Return _Context.NotNull AndAlso _Context.ChangeTracker.HasChanges()
      End Function

      Private Sub Test()
         Dim trk = _Context.ChangeTracker
         Dim X = trk.Entries.ToList
         Nop()
      End Sub

   End Class

End Namespace